/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#include <stdafx.h>
#include "brenderer.h"

#include <assert.h>

void base_renderer::clean() {
	if(_data) {
		delete[] _data;
		_data = 0;
		_width = _height = 0;
	}
}

base_renderer::base_renderer() : 
		_data(0), 
		_frame(0), 
		_nbframes(0), 
		_width(0), 
		_height(0) {
}

base_renderer::~base_renderer() { 
	clean(); 
}

bool base_renderer::init_frame(const point & p) {
	clean();
	_width = p.get_x();
	_height = p.get_y();
	assert(_width && _height);
	_data = new char[_width * _height];
	if(!_data) error("base_renderer unable to allocate frame buffer");
	return true;
}

char * base_renderer::lock_frame(int frame) {
	_frame = frame; 
	if(!_data) error("no allocated image buffer in lock_frame");
	return _data;
}

bool base_renderer::unlock_frame() {
	return true;
}

bool base_renderer::flip_frame() { 
	save(); 
	return true;
}

bool base_renderer::set_palette(const palette & pal) { 
	_pal = pal; 
	return true;
}
