/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#include <stdafx.h>
#include "channel.h"
#include "chunck.h"
#include "chunck_type.h"

#include <assert.h>
#include <string.h> // for memcpy.h
#ifndef min
#define min(x, y) ((x) > (y) ? (y) : (x))
#endif

#if 0

struct mem_line_list {
	int line;
	int what;
	struct mem_line_list * next;
};

struct mem_chunck {
	void * ptr;
	char * file;
	struct mem_line_list * lines;
	int size;
	int alloc;
	struct mem_chunck * next;
};

struct mem_chunck * s_my_chunck = 0;
	
void my_mem_info(void) {
	struct mem_chunck * m = s_my_chunck;
	//~ fprintf(stderr, "leak info : \n");
	while(m) {
		fprintf(stderr, "ALLOCATION of %d bytes\n", m->size);
		struct mem_line_list * lines = m->lines;
		int w;
		while(lines) {
			if(lines->what == 0) 
				fprintf(stderr, " allocated here : %d\n", lines->line);
			else if(lines->what == 1) 
				fprintf(stderr, " freed here : %d\n", lines->line);
			else
				fprintf(stderr, " changed here : %d\n", lines->line);
			w = lines->what;
			lines = lines->next;
		}
		if(w != 1) { fprintf(stderr, "ERROR_ON_THIS_ONE\n"); }
		m = m->next;
	}
}

void my_mem_changed_line(void * ptr, int line) {
	struct mem_chunck * m = s_my_chunck;
	while(m) {
		if(m->ptr == ptr) {
			struct mem_line_list * ptr = m->lines, * n;
			while(ptr->next) ptr = ptr->next;
			n = (struct mem_line_list *) malloc(sizeof(struct mem_line_list));
			n->line = line;
			n->what = 2;
			n->next = 0;
			ptr->next = n;
			return;
		}
		m = m->next;
	}
	warning("changing unknown memory at line %d", line);
};

unsigned char * my_alloc(int size, char * file, int line) {
	struct mem_chunck * m = (struct mem_chunck *) malloc(sizeof(struct mem_chunck));
	m->ptr = malloc(size);
	m->file = (char*)malloc(strlen(file)+1);
	strcpy(m->file, file);
	m->lines = (struct mem_line_list *) malloc(sizeof(struct mem_line_list));
	m->lines->line = line;
	m->lines->what = 0;
	m->lines->next = 0;
	m->size = size;
	if(s_my_chunck == 0) {
		atexit(my_mem_info);
	}
	m->next = s_my_chunck;
	s_my_chunck = m;
	return (unsigned char *) m->ptr;
}

void my_free(void * ptr, char * file, int line) {
	struct mem_chunck * m = s_my_chunck;
	while(m) {
		if(m->ptr == ptr) {
			struct mem_line_list * ptr = m->lines, * n;
			while(ptr->next) ptr = ptr->next;
			n = (struct mem_line_list *) malloc(sizeof(struct mem_line_list));
			n->line = line;
			n->what = 1;
			n->next = 0;
			ptr->next = n;
			return;
		}
		m = m->next;
	}
	warning("%s:%d freeing unknown memory", file, line);
}

#define ALLOC(x) my_alloc((x), __FILE__, __LINE__)
#define FREE(x) my_free((x), __FILE__, __LINE__)
#define CHANGE(x) my_mem_changed_line((x), __LINE__)

#endif
void saud_channel::handleStrk(chunck & b) {
	int size = b.get_size();
	if(size != 14 && size != 10) {
		error("STRK has a invalid size : %d", size);
	}
}

void saud_channel::handleSmrk(chunck & b) {
	_mark_reached = true;
}

void saud_channel::handleShdr(chunck & b) {
	int size = b.get_size();
	if(size != 4) warning("SMRK has a invalid size : %d", size);
}

bool saud_channel::handleSubTags(int & offset) {
	int available_size = _tbuffer_size - offset;
	if(available_size > 8) {
		chunck::type type = READ_BE_UINT32(_tbuffer + offset);
		unsigned int size = READ_BE_UINT32(_tbuffer + offset + 4);
		switch(type) {
			case TYPE_STRK:
				_in_data = false;
				if(available_size >= (size + 8)) {
					cont_chunck * c = new cont_chunck((char*)_tbuffer + offset);
					handleStrk(*c);
					delete c;
				} else return false;
				break;
			case TYPE_SMRK:
				_in_data = false;
				if(available_size >= (size + 8)) {
					cont_chunck * c = new cont_chunck((char*)_tbuffer + offset);
					handleSmrk(*c);
					delete c;
				} else return false;
				break;
			case TYPE_SHDR:
				_in_data = false;
				if(available_size >= (size + 8)) {
					cont_chunck * c = new cont_chunck((char*)_tbuffer + offset);
					handleShdr(*c);
					delete c;
				} else return false;
				break;
			case TYPE_SDAT: // Sound data !!!
				_in_data = true;
				_data_size = size;
				offset += 8;
				return false;
			default:
				error("unknown chunck in SAUD track : %s ", chunck::chunck_string(type));
		}
		offset += size + 8;
		return true;
	}
	return false;
}

bool saud_channel::processBuffer() {
	// At the start of this function, we have _tbuffer[0.._tbuffersize] containing possible data...
	// and _sbuffer is 0
	// At the end we have :
	// if(sound data) _sbuffer[0.._sbuffer_size] contains the sound data
	// the unprocessed data is kept in _tbuffer[0.._tbuffersize] (which may have changed)
	// if no unprocessed data, then _tbuffer is 0
	assert(_tbuffer != 0);
	assert(_tbuffer_size != 0);
	assert(_sbuffer == 0);
	assert(_sbuffer_size == 0);
	
	if(_in_data) {
		if(_data_size < _tbuffer_size) {
			// we assume that the channel is finished after data is received... (this assumption failed in realride.san)
			int offset= _data_size;
			while(handleSubTags(offset));
			_sbuffer_size = _data_size;
			_sbuffer = _tbuffer;
			if(offset < _tbuffer_size) { // there is still some unprocessed data
				int new_size = _tbuffer_size - offset;
				_tbuffer = new unsigned char[new_size];
				if(!_tbuffer)  error("saud_channel failed to allocate memory");
				memcpy(_tbuffer, _sbuffer + offset, new_size);
				_tbuffer_size = new_size;
			} else {
				_tbuffer = 0;
				_tbuffer_size = 0;
			}
			if(_sbuffer_size == 0) {
				warning("here !!");
				delete []_sbuffer; 
				_sbuffer = 0;
			}
		} else {
			// easy, swap the buffer
			_sbuffer_size = _tbuffer_size;
			_sbuffer = _tbuffer;
			_tbuffer_size = 0;
			_tbuffer = 0;
		}
	} else {
		int offset = 0;
		while(handleSubTags(offset));
		//~ assert(_in_data);
		if(_in_data) {
			//~ unsigned char * old = _tbuffer;
			_sbuffer_size = _tbuffer_size - offset;
			assert(_sbuffer_size);
			_sbuffer = new unsigned char[_sbuffer_size];
			if(!_sbuffer)  error("saud_channel failed to allocate memory");
			memcpy(_sbuffer, _tbuffer + offset, _sbuffer_size);
			delete []_tbuffer;
			_tbuffer = 0;
			_tbuffer_size = 0;
		} else {
			if(offset) { // maybe I should assert() this to avoid a lock...
				unsigned char * old = _tbuffer;
				int new_size = _tbuffer_size - offset;
				_tbuffer = new unsigned char[new_size];
				if(!_tbuffer)  error("saud_channel failed to allocate memory");
				memcpy(_tbuffer, old + offset, new_size);
				_tbuffer_size = new_size;
				delete []old;
			}
		}
	}
	return true;
}

saud_channel::saud_channel(int track, int freq) : 
			_track(track), 
			_nbframes(0),
			_data_size(-1),
			_tbuffer(0), 
			_sbuffer(0), 
			_frequency(freq),
			_tbuffer_size(0), 
			_sbuffer_size(0), 
			_in_data(false),
			_mark_reached(false) {
}

saud_channel::~saud_channel() {
	//~ warning("saud_channel::~saud_channel()");
	if(_tbuffer) delete []_tbuffer;
	if(_sbuffer) {
		warning("this should never happen !!!! (_sbuffer not NULL here)");
		delete []_sbuffer;
	}
	//~ my_mem_info();
}

bool saud_channel::is_terminated() const {
	//~ warning("saud2_channel::is_terminated() : %d, %d, %d", _track, _mark_reached, _data_size);
	return (_mark_reached && _data_size == 0 && _sbuffer == 0);
}

void saud_channel::recalcVolumeTable() {
	const int MAX_BALANCE = 100;
	int volume_left, volume_right;
	if(_balance < -MAX_BALANCE || _balance > MAX_BALANCE) {
		error("balance is out of range ! : %d", _balance);
	}
	int left_multiplier = MAX_BALANCE - _balance;
	int right_multiplier = MAX_BALANCE + _balance;
	volume_left = _volume * left_multiplier / (MAX_BALANCE*2);
	volume_right = _volume * right_multiplier / (MAX_BALANCE*2);
	if(volume_left < 0) volume_left = 0;
	if(volume_left > 128) volume_left = 128;
	if(volume_right < 0) volume_right = 0;
	if(volume_right > 128) volume_right = 128;
	for(int i = 0; i < 256; i++) {
		int value = volume_left * (signed char)i;
		_voltable[0][i] = TO_BE_16(value);
		value = volume_right * (signed char)i;
		_voltable[1][i] = TO_BE_16(value);
	}
}

bool saud_channel::set_parameters(int nb, int flags, int volume, int balance) {
	_nbframes = nb;
	_flags = flags; // bit 7 == IS_VOICE, bit 6 == IS_BACKGROUND_MUSIC, other ??
	_volume = volume;
	_balance = balance;
	_index = 0;
	recalcVolumeTable();
	return true;
}

bool saud_channel::check_parameters(int index, int nb, int flags, int volume, int balance) {
	if(++_index != index) error("invalid index in saud_channel::check_parameters()");
	if(_nbframes != nb) error("invalid duration in saud_channel::check_parameters()");
	if(_flags != flags) error("invalid flags in saud_channel::check_parameters()");
	if(_volume != volume || _balance != balance) {
		_volume = volume;
		_balance = balance;
		recalcVolumeTable();
	}
	return true;
}

bool saud_channel::append_data(chunck & b, int size) {
	if(_data_size == -1) { // First call
		assert(size > 8);
		chunck::type saud_type = b.get_dword(); saud_type = TO_BE_32(saud_type);
		unsigned int saud_size = b.get_dword(); saud_size = TO_BE_32(saud_size);
		if(saud_type != TYPE_SAUD) error("Invalid CHUNCK for saud_channel : %X", saud_type);
		size -= 8;
		_data_size = -2; // We don't get here again...
	}
	if(_tbuffer) {
		unsigned char * old = _tbuffer;
		_tbuffer = new unsigned char[_tbuffer_size + size];
		if(!_tbuffer)  error("saud_channel failed to allocate memory");
		memcpy(_tbuffer, old, _tbuffer_size);
		delete []old;
		b.read(_tbuffer + _tbuffer_size, size);
		_tbuffer_size += size;
	} else {
		_tbuffer_size = size;
		_tbuffer = new unsigned char[_tbuffer_size];
		if(!_tbuffer)  error("saud_channel failed to allocate memory");
		b.read(_tbuffer, _tbuffer_size);
	}
	return processBuffer();
}

int saud_channel::available_sound_data(void) const {
	return _sbuffer_size;
}

void saud_channel::get_sound_data(short * snd, int size) {
	for(int i = 0; i < size; i++) {
		snd[2*i] = _voltable[0][_sbuffer[i] ^ 0x80];
		snd[2*i+1] = _voltable[1][_sbuffer[i] ^ 0x80];
	}
	_data_size -= size;
	delete []_sbuffer;
	_sbuffer = 0;
	_sbuffer_size = 0;
}

imuse_channel::imuse_channel(int track, int freq) : 
			_track(track), 
			_buffer(0), 
			_buffer_size(0), 
			_frequency(freq), 
			_cur_buffer_write_pos(0),
			_cur_buffer_read_pos(0),
			_data_start(-1){
	_decoded_buffer = new short[_frequency*2]; // FIXME: This is too big
	assert(_decoded_buffer);
}

imuse_channel::~imuse_channel() {
	if(_buffer) { 
		delete []_buffer; 
	}
	delete []_decoded_buffer;
}

bool imuse_channel::is_terminated() const
{
	if(_cur_buffer_read_pos < (_buffer_size - 8)) return false; // Frame completely processed
	if(_data_start != -1 && available_sound_data() == 0) return true;
	return false;
}

bool imuse_channel::set_parameters(int nbframes, int size, int unk1, int unk2)
{
	_nbframes = nbframes;
	_size = size;
	_unk1 = unk1;
	_unk2 = unk2;
	_index = 0;
	return true;
}

bool imuse_channel::check_parameters(int index, int nbframes, int size, int unk1, int unk2)
{
	return true;
}

bool imuse_channel::append_data(chunck & b, int size)
{
	_size -= size;
	if(!_buffer) { // First call
		assert(size > 8);
		chunck::type imus_type = b.get_dword();
		imus_type = TO_BE_32(imus_type);
		unsigned int imus_size = b.get_dword();
		imus_size = TO_BE_32(imus_size);
		if(imus_type != TYPE_iMUS) error("Invalid CHUNCK for imuse_channel");
		size -= 8;
		_buffer_size = imus_size;
		assert(_buffer_size);
		_buffer = new unsigned char[_buffer_size];
		if(!_buffer)  error("imuse_channel failed to allocate memory");
		b.read(_buffer, size);
		_cur_buffer_write_pos = size;
	} else {
		assert(size <= (_buffer_size - _cur_buffer_write_pos));
		
		b.read(_buffer + _cur_buffer_write_pos, size);
		_cur_buffer_write_pos += size;
	}
	warning("done");
	return processBuffer();
}

bool imuse_channel::handleFormat(chunck & src) {
	if(src.get_size() != 20) error("invalid size for FRMT chunck");
	_imuse_start = TO_BE_32(src.get_dword());
	src.seek(4);
	_bitsize = TO_BE_32(src.get_dword());
	_rate = TO_BE_32(src.get_dword());
	_channels = TO_BE_32(src.get_dword());
	assert(_channels == 1 || _channels == 2);
	return true;
}

bool imuse_channel::handleText(chunck & src) {
	return true;
}

bool imuse_channel::handleRegion(chunck & src) {
	if(src.get_size() != 8) error("invalid size for REGN chunck");
	_imuse_length = TO_BE_32(src.get_dword());
	return true;
}

bool imuse_channel::handleStop(chunck & src) {
	if(src.get_size() != 4) error("invalid size for STOP chunck");
	return true;
}

bool imuse_channel::handleMap(chunck & map) {
	while(!map.eof()) {
		chunck * sub = map.sub_block();
		//~ if(sub.get_size() & 1) map.seek(1);
		switch(sub->get_type()) {
			case TYPE_FRMT:
				handleFormat(*sub);
				break;
			case TYPE_TEXT:
				handleText(*sub);
				break;
			case TYPE_REGN:
				handleRegion(*sub);
				break;
			case TYPE_STOP:
				handleStop(*sub);
				break;
			default:
				error("Unknown iMUS subchunck found : %s, %d", chunck::chunck_string(sub->get_type()), sub->get_size());
		}
		delete sub;
	}
	return true;
}

int imuse_channel::decode(int size, int &ret)
{
	short * decoded = _decoded_buffer;
	unsigned char * source = _buffer + _data_start;
	assert(_bitsize == 12);
	int msize = size & 1;
	size = (size+1) >> 1;
	while(size--) {
		int v1 =  *source++;
		int v2 =  *source++;
		int v3 =  *source++;
		int value = (((v2 & 0x0f) << 12) | (v1 << 4)) - 0x8000;
		*decoded++ = (short)value;
		value = (((v2 & 0xf0) << 8) | (v3 << 4)) - 0x8000;		
		*decoded++ = (short)value;
	} 
	if(msize) decoded--;
	ret = source - (_buffer + _data_start);
	return decoded - _decoded_buffer;
}

bool imuse_channel::processBuffer() {
	while((_cur_buffer_write_pos - _cur_buffer_read_pos) >= 8) {
		int available_size = _cur_buffer_write_pos - _cur_buffer_read_pos;
		chunck::type type = READ_BE_UINT32(_buffer + _cur_buffer_read_pos);
		unsigned int size = READ_BE_UINT32(_buffer + _cur_buffer_read_pos + 4);
		switch(type) {
			case TYPE_MAP_: 
				{
					assert(available_size >= (size + 8));
					cont_chunck map_chunck((char*)_buffer + _cur_buffer_read_pos);
					handleMap(map_chunck);
				} break;
			case TYPE_DATA: // Sound data !!!
				if(_data_start == -1) {
					_rdata_start = _data_start = _cur_buffer_read_pos + 8;
					_rdata_size = _data_size = size;
					int reqsize = 1;
					if(_channels == 2) reqsize *= 2;
					if(_bitsize == 16) reqsize *= 2;
					else if(_bitsize == 12) reqsize *= 3;
					if((size % reqsize) != 0)
						error("Invalid sound data size");
				}
				break;
			case TYPE_iMUS:
				size = 0;
				break;
			default:
				error("unknown chunck in iMUS track : %s", chunck::chunck_string(type));
				break;
		}
		if(available_size >= (size + 8)) {
			_cur_buffer_read_pos += size + 8;
		} else {
			break;
		}
	}
	return true;
}

int imuse_channel::available_sound_data(void) const {
	assert(_data_start <= _cur_buffer_write_pos);
	int size = min(_cur_buffer_write_pos - _data_start, _data_size);
	
	switch(_bitsize) {
		case 12:
			size = (size / 3) * 2;
			break;
		case 16:
			size /= 2;
	}
	if(_channels == 2)
		size /= 2;
	if(_bitsize == 12) size &= ~1;
	return size;
}

void imuse_channel::get_sound_data(short * snd, int size) {
	if(_data_start == -1 || _data_size == 0 || _bitsize <= 8) error("invalid call to saud_channel::read_sound_data()");
#ifdef DEBUG_CHANNEL
	int available =  min(_cur_buffer_write_pos - _data_start, _data_size) * 2;
	switch(_bitsize) {
		case 12:
			available = (available / 3) * 2;
			break;
		case 16:
			available /= 2;
	}
	assert(available >= size);
#endif
	int really_decoded = 0;
	if(_bitsize == 12) {
		int decoded_size = decode((_channels == 2) ? size*2 : size, really_decoded);
	}
	_data_start += really_decoded;
	_data_size -= really_decoded;
	if(_channels == 2) size *= 2;
	for(int i = 0; i < size * 2; i++)
		snd[i] += _decoded_buffer[i];
}

void imuse_channel::get_sound_data(char * snd, int size) {
	if(_data_start == -1 || _data_size == 0 || _bitsize > 8) error("invalid call to saud_channel::read_sound_data()");
	
#ifdef DEBUG_CHANNEL
	int available =  min(_cur_buffer_write_pos - _data_start, _data_size) * 2;
	assert(available >= size);
#endif
	_data_start += size;
	_data_size -= size;
	if(_channels == 2) size *= 2;
	for(int i = 0; i < size * 2; i++)
		snd[i] += _decoded_buffer[i];
}



