/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#include <stdafx.h>
#include "channel.h"
#include "chunck.h"
#include "chunck_type.h"

#include <assert.h>
#include <string.h> // for memcpy.h
#ifndef min
#define min(x, y) ((x) > (y) ? (y) : (x))
#endif

imuse_channel::imuse_channel(int track, int freq) : 
			_track(track), 
			_tbuffer(0), 
			_tbuffer_size(0), 
			_sbuffer(0), 
			_sbuffer_size(0), 
			_frequency(freq), 
			_data_size(-1),
			_in_data(false)
{
}

imuse_channel::~imuse_channel() {
	//~ warning("%p : ~imuse_channel(%d) : %p(%d), %p(%d, %d)", this, _track, _data_size, _tbuffer, _tbuffer_size, _sbuffer, _sbuffer_size, _srbuffer_size);
	if(_tbuffer) {
		delete []_tbuffer; 
	}
	if(_sbuffer) {
		warning("_sbuffer should be 0 !!!");
		delete []_sbuffer; 
	}
}

bool imuse_channel::is_terminated() const
{
	//~ warning("%p : is_terminated(%d, %d) : %p(%d), %p(%d, %d)", this, _track, _data_size, _tbuffer, _tbuffer_size, _sbuffer, _sbuffer_size, _srbuffer_size);
	return (_data_size <= 0 && _sbuffer == 0);
}

bool imuse_channel::set_parameters(int nbframes, int size, int unk1, int unk2)
{
	//~ _nbframes = nbframes;
	//~ _size = size;
	//~ _unk1 = unk1;
	//~ _unk2 = unk2;
	//~ _index = 0;
	return true;
}

bool imuse_channel::check_parameters(int index, int nbframes, int size, int unk1, int unk2)
{
	return true;
}

bool imuse_channel::append_data(chunck & b, int size)
{
	//~ _size -= size;
	if(_data_size == -1) { // First call
		assert(size > 8);
		chunck::type imus_type = b.get_dword(); imus_type = TO_BE_32(imus_type);
		unsigned int imus_size = b.get_dword(); imus_size = TO_BE_32(imus_size);
		if(imus_type != TYPE_iMUS) error("Invalid CHUNCK for imuse_channel");
		size -= 8;
		_tbuffer_size = size;
		assert(_tbuffer_size);
		_tbuffer = new unsigned char[_tbuffer_size];
		if(!_tbuffer)  error("imuse_channel failed to allocate memory");
		b.read(_tbuffer, size);
		_data_size = -2; // even if _in_data does not get set, this won't be called again
	} else {
		if(_tbuffer) { // remaining from last call
			unsigned char * old = _tbuffer;
			int new_size = size + _tbuffer_size;
			_tbuffer = new unsigned char[new_size];
			if(!_tbuffer)  error("imuse_channel failed to allocate memory");
			memcpy(_tbuffer, old, _tbuffer_size);
			delete []old;
			b.read(_tbuffer + _tbuffer_size, size);
			_tbuffer_size += size;
		} else {
			_tbuffer_size = size;
			_tbuffer = new unsigned char[_tbuffer_size];
			if(!_tbuffer)  error("imuse_channel failed to allocate memory");
			b.read(_tbuffer, size);
		}
	}
	return processBuffer();
}

bool imuse_channel::handleFormat(chunck & src) {
	if(src.get_size() != 20) error("invalid size for FRMT chunck");
	unsigned imuse_start = src.get_dword();
	imuse_start = TO_BE_32(imuse_start);
	src.seek(4);
	_bitsize = src.get_dword();
	_bitsize = TO_BE_32(_bitsize);
	_rate = src.get_dword();
	_rate = TO_BE_32(_rate);
	_channels = src.get_dword();
	_channels = TO_BE_32(_channels);
	assert(_channels == 1 || _channels == 2);
	return true;
}

bool imuse_channel::handleText(chunck & src) {
	return true;
}

bool imuse_channel::handleRegion(chunck & src) {
	if(src.get_size() != 8) error("invalid size for REGN chunck");
	return true;
}

bool imuse_channel::handleStop(chunck & src) {
	if(src.get_size() != 4) error("invalid size for STOP chunck");
	return true;
}

bool imuse_channel::handleMap(chunck & map) {
	while(!map.eof()) {
		chunck * sub = map.sub_block();
		//~ warning("%s, %d", chunck::chunck_string(sub->get_type()), sub->get_size());
		//~ if(sub.get_size() & 1) map.seek(1);
		switch(sub->get_type()) {
			case TYPE_FRMT:
				handleFormat(*sub);
				break;
			case TYPE_TEXT:
				handleText(*sub);
				break;
			case TYPE_REGN:
				handleRegion(*sub);
				break;
			case TYPE_STOP:
				handleStop(*sub);
				break;
			default:
				error("Unknown iMUS subchunck found : %s, %d", chunck::chunck_string(sub->get_type()), sub->get_size());
		}
		delete sub;
	}
	return true;
}

void imuse_channel::decode()
{
	int remaining_size = _sbuffer_size % 3;
	if(remaining_size) {
		//~ warning("%d remaining byte(s) in 12bit iMUS stream, this is boring", remaining_size);
		_srbuffer_size -= remaining_size;
		assert(_in_data);
		if(_tbuffer == 0) {
			_tbuffer = new unsigned char[remaining_size];
			memcpy(_tbuffer, _sbuffer + _sbuffer_size - remaining_size, remaining_size);
			_tbuffer_size = remaining_size;
			_sbuffer_size -= remaining_size;
		} else {
			warning("impossible ! : %p, %d, %d, %p(%d), %p(%d, %d)",
				this, _data_size, _in_data, _tbuffer, _tbuffer_size, _sbuffer, _sbuffer_size, _srbuffer_size);
			unsigned char * old = _tbuffer;
			int new_size = remaining_size + _tbuffer_size;
			_tbuffer = new unsigned char[new_size];
			if(!_tbuffer)  error("imuse_channel failed to allocate memory");
			memcpy(_tbuffer, old, _tbuffer_size);
			delete []old;
			memcpy(_tbuffer + _tbuffer_size, _sbuffer + _sbuffer_size - remaining_size, remaining_size);
			_tbuffer_size += remaining_size;
		}
	}
	int loop_size = _sbuffer_size / 3;
	int new_size = loop_size * 2;
	short * keep, * decoded;
	keep = decoded = new short[new_size];
	assert(keep);
	unsigned char * source = _sbuffer;
	while(loop_size--) {
		int v1 =  *source++;
		int v2 =  *source++;
		int v3 =  *source++;
		int value = (((v2 & 0x0f) << 12) | (v1 << 4)) - 0x8000;
		*decoded++ = (short)value;
		value = (((v2 & 0xf0) << 8) | (v3 << 4)) - 0x8000;		
		*decoded++ = (short)value;
	}
	delete []_sbuffer;
	_sbuffer = (unsigned char*)keep;
	_sbuffer_size = new_size * sizeof(short);
	//~ debug(7, "new_size == %d, sbuffer_size == %d\n", new_size, _sbuffer_size);
}


bool imuse_channel::handleSubTags(int & offset) {
	int available_size = _tbuffer_size - offset;
	if(available_size >= 8) {
		chunck::type type = READ_BE_UINT32(_tbuffer + offset);
		unsigned int size = READ_BE_UINT32(_tbuffer + offset + 4);
		switch(type) {
			case TYPE_MAP_: 
				_in_data = false;
				if(available_size >= (size + 8)) {
					cont_chunck c((char*)_tbuffer + offset);
					handleMap(c);
				} break;
			case TYPE_DATA: // Sound data !!!
				_in_data = true;
				_data_size = size;
				offset += 8;
				{
					int reqsize = 1;
					if(_channels == 2) reqsize *= 2;
					if(_bitsize == 16) reqsize *= 2;
					else if(_bitsize == 12) { if(reqsize > 1) reqsize = reqsize * 3 / 2; else reqsize = 3; }
					if((size % reqsize) != 0) {
						warning("Invalid iMUS sound data size : (%d %% %d) != 0, correcting...", size, reqsize);
						size += 3 - (size % reqsize);
					}
				} return false;
			default:
				error("unknown chunck in iMUS track : %s ", chunck::chunck_string(type));
		}
		offset += size + 8;
		return true;
	}
	return false;
}

bool imuse_channel::processBuffer() {
	// see comments in saud_channel::processBuffer for an explanation
	assert(_tbuffer != 0);
	assert(_tbuffer_size != 0);
	assert(_sbuffer == 0);
	assert(_sbuffer_size == 0);
	
	if(_in_data) {
		if(_data_size < _tbuffer_size) {			
			int offset= _data_size;
			while(handleSubTags(offset));
			_sbuffer_size = _data_size;
			_sbuffer = _tbuffer;
			if(offset < _tbuffer_size) { // there is still some unprocessed data
				int new_size = _tbuffer_size - offset;
				_tbuffer = new unsigned char[new_size];
				if(!_tbuffer)  error("imuse_channel failed to allocate memory");
				memcpy(_tbuffer, _sbuffer + offset, new_size);
				_tbuffer_size = new_size;
			} else {
				_tbuffer = 0;
				_tbuffer_size = 0;
			}
			if(_sbuffer_size == 0) {
				// this never happened yet, but who knows
				delete []_sbuffer; 
				_sbuffer = 0;
			}
		} else {
			// easy, swap the buffer
			_sbuffer_size = _tbuffer_size;
			_sbuffer = _tbuffer;
			_tbuffer_size = 0;
			_tbuffer = 0;
		}
	} else {
		int offset = 0;
		while(handleSubTags(offset));
		//~ assert(_in_data);
		if(_in_data) {
			//~ unsigned char * old = _tbuffer;
			_sbuffer_size = _tbuffer_size - offset;
			assert(_sbuffer_size);
			_sbuffer = new unsigned char[_sbuffer_size];
			if(!_sbuffer)  error("imuse_channel failed to allocate memory");
			memcpy(_sbuffer, _tbuffer + offset, _sbuffer_size);
			delete []_tbuffer;
			_tbuffer = 0;
			_tbuffer_size = 0;
		} else {
			if(offset) { // maybe I should assert() this to avoid a lock...
				unsigned char * old = _tbuffer;
				int new_size = _tbuffer_size - offset;
				_tbuffer = new unsigned char[new_size];
				if(!_tbuffer)  error("imuse_channel failed to allocate memory");
				memcpy(_tbuffer, old + offset, new_size);
				_tbuffer_size = new_size;
				delete []old;
			}
		}
	}
	_srbuffer_size = _sbuffer_size;
	if(_sbuffer && _bitsize == 12) decode();
	return true;
}

int imuse_channel::available_sound_data(void) const {
	int ret = _sbuffer_size;
	if(_channels == 2) ret /= 2;
	if(_bitsize > 8) ret /= 2;
	//~ debug(7, "buffer_size == %d, ret == %d\n", _sbuffer_size, ret);
	return ret;
}

void imuse_channel::get_sound_data(short * snd, int size) {
	if(_data_size <= 0 || _bitsize <= 8) error("invalid call to imuse_channel::read_sound_data()");
	if(_channels == 2) size *= 2;
	for(int i = 0; i < size; i++)
		snd[i] = READ_BE_UINT16(_sbuffer+2*i);
	delete []_sbuffer;
	assert(_sbuffer_size == 2*size);
	_sbuffer = 0;
	_sbuffer_size = 0;
	_data_size -= _srbuffer_size;
}

void imuse_channel::get_sound_data(char * snd, int size) {
	if(_data_size <= 0 || _bitsize > 8) error("invalid call to imuse_channel::read_sound_data()");
	if(_channels == 2) size *= 2;
	for(int i = 0; i < size; i++)
		snd[i] = _sbuffer[i];
	delete []_sbuffer;
	_sbuffer = 0;
	_sbuffer_size = 0;
	_data_size -= _srbuffer_size;
}
