/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#ifndef __MIXER_H_
#define __MIXER_H_

#include "config.h"

#ifdef DEBUG
# ifndef NO_DEBUG_MIXER
#  define DEBUG_MIXER
# endif
#else
# ifdef DEBUG_MIXER
#  error DEBUG_MIXER defined without DEBUG
# endif
#endif

class channel;

class sound_renderer;

/*! 	@brief The class for the player's sound mixer

	This class is used for sound mixing.
	It contains a list of current track and request them to mix.
	It then sends the mixed sound samples to the sound renderer.

*/
class mixer {
public:
	virtual ~mixer() {};
	virtual bool init() = 0;
	virtual channel * find_channel(int track) = 0;
	virtual bool add_channel(channel * c) = 0;
	virtual bool handle_frame() = 0;
	virtual bool stop() = 0;
};

#if 0
// Xavier : This is my old mixer class, not needed anymore

/*! 	@brief interface for sound output

	This is the interface for sound output. Only one implementation is available right now : ds_wave_out
*/
class sound_renderer {
public:
	virtual ~sound_renderer() {};
	/*!	@brief sound output initialisation
		
		This is called by the mixer when initialisation has to be performed.
		
		@param frequency is always 22050 right now.
		@param stereo is true if the sound output should be stereo (always true right now)

		@return true if initialisation was ok, false otherwise
	*/
	virtual bool init(int frequency, bool stereo) = 0;
	/*!	@brief sound data output
		
		This is called by the mixer to put data to the output.
		
		@param ptr buffer containing samples.
		@param size size of the \c ptr buffer, in samples (i.e. size in byte / 2)

		@return true if everything went fine, false otherwise
	*/
	virtual bool write(unsigned short * ptr, int size) = 0;
	/*!	@brief stop sound output
		
		This is called by the mixer when sound output should be stopped.

		@return true if everything went fine, false otherwise
	*/
	virtual bool stop() = 0;
};

class internal_mixer : public mixer {
private:
	static const int MAX_CHANNELS = 32;
	struct {
		int id;
		channel * chan;
	} _channels[MAX_CHANNELS];		//!< The map of track and channels
	int _frequency;					//!< The mixer's frequency
	short * _sound;					//!< Internal mixed sound buffer
public:
	internal_mixer(sound_renderer *);
	virtual ~mixer();
	bool init();
	channel * find_channel(int track);
	bool add_channel(channel * c);
	bool handle_frame();
	bool stop();
};
#endif // 0

#endif
