/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#ifndef __PLAYER_H_
#define __PLAYER_H_

#include "config.h"

#include "rect.h"
#include "mixer.h"
#include "chunck.h"
#include "palette.h"
#include "codec1.h"
#include "codec37.h"
#include "codec44.h"
#include "codec47.h"
#include "frenderer.h"

class renderer;

class string_resource;

/*!	@brief the SMUSH player class

	This class is the player itself. 
*/
class smush_player {
private:
	char * _fname;			//!< the name of the animation file being played
	int _version;			//!< the version of the animation file being played
	int _secondary_version;	//!< the secondary version number of the animation file being played
	int _sound_frequency;		//!< the sound frequency of the animation file being played
	int _nbframes;			//!< the number of frames in the animation file
	mixer * _mixer;			//!< the sound mixer
	palette _pal;			//!< the current palette
	short _delta_pal[768];		//!< the delta palette information set by an xpal
	renderer * _renderer;		//!< pointer to the ::renderer
	string_resource * _strings;	//!< pointer to the string resources associated with the animation
	font_renderer * _fr[4];		//!< pointers to the fonts for the animation
	codec1_decoder _codec1;	//!< the ::decoder for codec 1 and 3
	codec37_decoder _codec37;	//!< the ::decoder for codec 37
	codec44_decoder _codec44;	//!< the ::decoder for codec 21 and 44
	dump_decoder _codecd;	//!< the ::decoder for codec 21 and 44
	point _frame_size;		//!< the current frame size of the animation
	int _frame;				//!< the current frame number of the animation
	bool _output_sound;		//!< should we handle sound ?
	bool _wait;				//!< should we synchronise the player ?
	bool _already_init;		//!< has the player already been initialized for the current frame
	bool _codec37_called;		//!< has the codec 37 already been called once for this animation
	bool _skipnext;			//!< should the player skip the next frame object ?
	bool _subtitles;			//!< should the player handle subtitles ?
	bool _bgmusic;			//!< should the player output the background music ?
	bool _voices;			//!< should the player output the voice ?
	bool _skips[37];			//!< mapping of frame object identifier to show or hide
	char * _cur_buffer;		//!< pointer to the current frame
public:
	smush_player(renderer *, bool wait = true, bool output_sound = true);
	virtual ~smush_player();
	bool play(const char *);
	void update_palette(void);
	void show(const char *);
	void hide(const char *);
protected:
	bool read_string(const char * file, bool &);
	void clean();
	void checkBlock(const chunck &, chunck::type, unsigned int = 0);
	void handleAnimHeader(chunck &);
	void handleFrame(chunck &);
	void handleNewPalette(chunck &);
	void handleFrameObject(chunck &);
	void handleSoundBuffer(int, int, int, int, int, int, chunck &, int);
	void handleImuseBuffer(int, int, int, int, int, int, chunck &, int);
	void handleSoundFrame(chunck &);
	void handleSkip(chunck &);
	void handleStore(chunck &);
	void handleFetch(chunck &);
	void handleImuseAction8(chunck &, int flags, int unknown, int track_id);
	void handleImuseAction(chunck &);
	void handleTextResource(chunck &);
	void handleDeltaPalette(chunck &);
	void decodeCodec(chunck &, const rect &, decoder &);
	void readPalette(palette &, chunck &);
	void init_size(const rect &, bool, bool);
};

#endif
