/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#include <stdafx.h>
#include "rect.h"

rect::rect() : _topleft(0, 0), _bottomright(0,0) { 
}

rect::rect(int x, int y) : _topleft(0, 0), _bottomright(x, y) {
	check();
}

rect::rect(int x1, int y1, int x2, int y2) : _topleft(x1, y1), _bottomright(x2, y2) { 
	check(); 
}

rect::rect(const rect & r) : _topleft(r._topleft), _bottomright(r._bottomright) {
}

rect::rect & rect::operator=(const rect & r)  {
	_topleft = r._topleft;
	_bottomright = r._bottomright;
	return *this;
}

bool rect::operator==(const rect & r) const {
	return _topleft == r._topleft && _bottomright == r._bottomright;
}

void rect::check() { 
	if ((_topleft.get_x() < 0) || (_bottomright.get_x() < _topleft.get_x())  || (_topleft.get_y() < 0) || (_bottomright.get_y() < _topleft.get_y()))
	{
		error("Invalid rect");
	}
}

bool rect::is_inside(int x, int y) const {
	return _topleft.get_x() >= x && _bottomright.get_x() < x && _topleft.get_y() >= y && _bottomright.get_y() < y;
}

bool rect::is_inside(const point & p) const {
	return (left() <= p.get_x()) && (right() > p.get_x()) && (top() <= p.get_y()) && (bottom() > p.get_y());
}

